@@ -8,6 +8,7 @@ use v5.20;
 use File::HomeDir;
 use Path::Tiny;
 use Config::Model;
+use Config::Model::Lister;
 use Log::Log4perl qw(get_logger :levels);
 use YAML::PP;
 
@@ -17,8 +18,6 @@ use App::Cme -command ;
 
 use base qw/App::Cme::Common/;
 use feature qw/postderef signatures/;
-no warnings qw/experimental::postderef experimental::signatures experimental::smartmatch/;
-
 
 my $__test_home = '';
 # used only by tests
@@ -39,6 +38,8 @@ sub opt_spec {
     return ( 
         [ "arg=s@"  => "script argument. run 'cme run <script> -doc' for possible arguments" ],
         [ "backup:s"  => "Create a backup of configuration files before saving." ],
+        [ "foreach=s" => "Run script in several directories. The list of directories "
+          . "must be passed as a single argument, i.e. --foreach 'foo bar'" ],
         [ "commit|c:s" => "commit change with passed message" ],
         [ "cat" => "Show the script file" ],
         [ "no-commit|nc!" => "skip commit to git" ],
@@ -71,10 +72,15 @@ sub check_script_arguments ($self, $opt, $script_name) {
         my @scripts;
         foreach my $path ( @script_paths ) {
             next unless $path->is_dir;
-            push @scripts, grep { ! /~$/ } $path->children();
+            push @scripts, map {$_->basename} grep { ! /~$/ } $path->children();
         }
         say $opt->{list} ? "Available scripts:" : "Missing script argument. Choose one of:";
-        say map {"- ".$_->basename."\n"} @scripts ;
+        foreach my $script_path (sort @scripts) {
+            my $data = $self->get_script_data($script_path);
+            printf("- %s (app: %s)\n",$script_path, $data->{app} );
+        }
+        say "";
+        say "Run 'cme run <script> -doc' to get more details on a script.";
         return 0;
     }
     return 1;
@@ -183,6 +189,7 @@ sub parse_script_lines ($script, $lines) {
 }
 
 sub parse_command ($key, $parsed_data, $line_nb, $value) {
+    ## no critic (ControlStructures::ProhibitCascadingIfElse)
     if ( $key =~ /^app/ ) {
         $parsed_data->{app} = $value->[0];
     }
@@ -264,9 +271,10 @@ sub parse_script ($script, $content, $user_args) {
         return $data;
     }
 
+    my $data;
     if ($lines->[0] =~ /Format:\s*yaml/i) {
         my $ypp = YAML::PP->new;
-        my $data = $ypp->load_string($content);
+        $data = $ypp->load_string($content);
         foreach my $key (qw/doc code load var/) {
             next unless defined $data->{$key};
             next if ref $data->{$key} eq 'ARRAY';
@@ -277,19 +285,54 @@ sub parse_script ($script, $content, $user_args) {
         if ($data->{default} and ref $data->{default} ne 'HASH') {
             die "default spec must be a hash ref, not a ", ref $data->{default} // 'scalar', "\n";
         }
-        $data = process_script_vars ($user_args, $data);
-        return $data;
+    }
+    else {
+        $data = parse_script_lines ($script, $lines);
     }
 
-    my $data = parse_script_lines ($script, $lines);
-    $data = process_script_vars ($user_args, $data);
-    return $data;
+    my $processed_data = process_script_vars ($user_args, $data);
+
+    if (not $processed_data->{app}) {
+        die "Missing 'app' parameter in script $script.\n"
+    }
+
+    return $processed_data;
 }
 
 sub commit ($self, $root, $msg) {
     $msg =~ s/\{\{(.*?)\}\}/$root->grab_value($1)/e;
 
-    system(qw/git commit -a -m/, $msg);
+    system(qw/git commit -a -m/, $msg) == 0
+        or die "git commit failed: $?\n";
+
+    return;
+}
+
+# returns: script file name, script data if script is *not* Perl code
+sub get_script_data ($self, $script_name, $opt = {}) {
+    my $script_file = $self->find_script_file($script_name);
+
+    my $content = $script_file->slurp_utf8;
+
+    if ($content =~ m/^#!/ and -x $script_file) {
+        return $script_file, {};
+    }
+
+    # parse variables passed on command line
+    my %user_args = map { split '=',$_,2; } @{ $opt->{arg} };
+
+    return ($script_file, parse_script($script_file, $content, \%user_args));
+}
+
+sub run_script_as_code ($self, $script_name, $script_file) {
+    splice @ARGV, 0,2;          # remove 'run script' arguments
+    my $done = eval $script_file->slurp_utf8."\n1;\n"; ## no critic (BuiltinFunctions::ProhibitStringyEval)
+    if (ref $done eq 'HASH') {
+        warn "script $script_name returns a hash but it's processed as a plain script.",
+            " This may not be what you want\n";
+    }
+    die "Error in script $script_name: $@\n" unless $done;
+    return;
 }
 
 sub execute {
@@ -304,30 +347,18 @@ sub execute {
 
     return unless $self->check_script_arguments($opt, $script_name);
 
-    my $script = $self->find_script_file($script_name);
-
-    my $content = $script->slurp_utf8;
+    my ($script_file, $script_data) = $self->get_script_data($script_name, $opt);
 
     if ($opt->{cat}) {
-        print $content;
+        print $script_file->slurp_utf8;
         return;
     }
 
-    # parse variables passed on command line
-    my %user_args = map { split '=',$_,2; } @{ $opt->{arg} };
-
-    if ($content =~ m/^#!/ and -x $script) {
-        splice @ARGV, 0,2; # remove 'run script' arguments
-        my $done = eval $script->slurp_utf8."\n1;\n"; ## no critic (BuiltinFunctions::ProhibitStringyEval)
-        if (ref $done eq 'HASH') {
-            warn "script $script_name returns a hash but it's processed as a plain script.",
-                " This may not be what you want\n";
-        }
-        die "Error in script $script_name: $@\n" unless $done;
+    if (not defined $script_data->{app}) {
+        $self->run_script_as_code ($script_name, $script_file);
         return;
     }
 
-    my $script_data = parse_script($script, $content, \%user_args);
     my $commit_msg = $script_data->{commit_msg};
 
     if ($opt->doc) {
@@ -337,7 +368,8 @@ sub execute {
     }
 
     if (my @missing = sort keys $script_data->{missing}->%*) {
-        die "Error: Missing variables '". join("', '",@missing)."' in command arguments for script $script\n"
+        die "Error: Missing variables '". join("', '",@missing)
+            ."' in command arguments for script $script_file\n"
             ."Please use option '".join(' ', map { "-arg $_=xxx"} @missing)."'\n";
     }
 
@@ -349,16 +381,75 @@ sub execute {
         $commit_msg = $opt->{commit};
     }
 
+    $opt->{_verbose} = 'Loader' if $opt->{verbose};
+
+    my ( $categories, $appli_info, $appli_map ) = Config::Model::Lister::available_models;
+    my $app = $script_data->{app};
+    if ($opt->{foreach} and $appli_info->{$app}{_category} ne 'application') {
+        die "Cannot use --foreach option with $app. This option can only be used with ".
+            join(' ', $categories->{application}->@*). ". Check your cme script.\n";
+    }
+
+    # parse variables passed on command line
+
+    if ($opt->{foreach}) {
+        $self->run_foreach_loop($opt,$app_args, $script_data);
+    }
+    else {
+        my %user_args = map { split '=',$_,2; } @{ $opt->{arg} };
+        $self->run_script ($opt, $app_args, $script_data, \%user_args);
+    }
+
+    return;
+}
+
+sub run_foreach_loop($self, $opt,$app_args, $script_data ) {
+    my %user_args = map { split '=',$_,2; } @{ $opt->{arg} };
+
+    my @dirs = map { chomp ; split /\s+/; }
+        ($opt->{foreach} eq '-' ? <STDIN> : ($opt->{foreach}));
+
+    my $start = path('.')->absolute;
+
+    foreach my $d (@dirs) {
+        my $t_dir = $start->child($d);
+        if (not $t_dir->is_dir) {
+            die "Cannot run script in $d: not a directory\n";
+        }
+        say "Running script in $t_dir ...";
+        # instance is stored in Config::Model, so the name must be changed
+        $opt->{instance_name} = $d;
+        # instance is persisted in $self, so its ref must be removed
+        delete $self->{_instance};
+        chdir $t_dir->stringify;
+        $self->run_script ($opt, $app_args, $script_data, {%user_args});
+        # once we're done, remove instance from Model to avoid memory leaks
+        # TODO: use delete_instance method provided by Config::Model from version 2.156
+        delete $self->{_model}{instances}{$d};
+    }
+
+    chdir $start->stringify;
+    return;
+}
+
+sub run_script ($self, $opt, $app_args, $script_data, $user_args){
+    my $commit_msg = $script_data->{commit_msg};
+    my $stashed;
+
     # check if workspace and index are clean
     if ($commit_msg and not $opt->{no_commit}) {
         ## no critic(InputOutput::ProhibitBacktickOperators)
         my $r = `git status --porcelain --untracked-files=no`;
-        die "Cannot run commit command in a non clean repo. Please commit or stash pending changes: $r\n"
-            if $r;
+        if ($?) {
+            die "git status command failed: $?\n";
+        }
+        if ($r) {
+            system(qw/git stash push --quiet --message/, "cme run auto stash") == 0
+                or die "git stash push failed: $?\n";
+            $stashed = 1;
+        };
     }
 
-    $opt->{_verbose} = 'Loader' if $opt->{verbose};
-
     # call loads
     my ($model, $inst, $root) = $self->init_cme($opt,$app_args);
     foreach my $load_str ($script_data->{load}->@*) {
@@ -376,19 +467,24 @@ sub execute {
     }
 
     if ($script_data->{sub}) {
-        $script_data->{sub}->($root, \%user_args);
+        $script_data->{sub}->($root, $user_args);
     }
 
-    unless ($inst->needs_save) {
+    if ($inst->needs_save) {
+        $self->save($inst,$opt) ;
+
+        # commit if needed
+        if ($commit_msg and not $opt->{no_commit}) {
+            $self->commit($root, $commit_msg);
+        }
+    } else {
         say "No change were applied";
-        return;
     }
 
-    $self->save($inst,$opt) ;
 
-    # commit if needed
-    if ($commit_msg and not $opt->{no_commit}) {
-        $self->commit($root, $commit_msg);
+    if ($stashed) {
+        system(qw/git stash pop --quiet/) == 0
+            or die "git stash pop failed: $?\n";
     }
 
     return;
@@ -448,6 +544,11 @@ __END__
  - update-copyright
  - add-me-to-uploaders
 
+ # mass modification
+ $ cme run add-me-to-uploaders --foreach "raku-log raku-zef"
+ # similar result
+ $ echo "raku-*" | cme run add-me-to-uploaders --foreach -
+
 =head1 DESCRIPTION
 
 Run a script written for C<cme>
@@ -579,8 +680,9 @@ Specify Perl code to run. See L</code section> for details.
 =item commit
 
 Specify that the change must be committed with the passed commit
-message. When this option is used, C<cme> raises an error if used on a
-non-clean workspace. This option works only with L<git>.
+message. When this option is used, C<cme> stashes and restores all
+modifications if used on a non-clean workspace. This option works only
+with L<git>.
 
 Strings like C<{{ load path }}> are substituted with a value extracted
 from configuration tree with the specified load path. See
@@ -731,6 +833,21 @@ committed with the passed commit message.
 
 Don't commit to git (even if the above option is set)
 
+=head2 foreach
+
+Apply the script in specific directories. This option is available
+only when dealing with application (See the list returned by C<cme list>).
+
+For instance, if your directory contains several Debian packages and
+you want to apply the same modification to all packages, you can run
+something like:
+
+   cme run add-me-to-uploaders --foreach "raku-log raku-zef"
+
+or use C<STDIN> to send the package directories:
+
+   ls "raku-*" | cme run add-me-to-uploaders --foreach -
+
 =head2 verbose
 
 Show effect of the modify instructions.
@@ -809,6 +926,14 @@ This script can also be written using multi line instructions:
  load: ! source Vcs-Browser="$browser" Vcs-Git="$url"
  commit: control: update Vcs-Browser and Vcs-Git
 
+=head2 personal scripts
+
+You may create your own scripts in C<~/.cme/scripts>. Here's an example:
+
+ $ cat ~/.cme/scripts/set-any-arch
+ app: dpkg-control
+ load: binary:~".*" Architecture=any
+
 =head1 SEE ALSO
 
 L<cme>
